from typing import List, Union
import numpy as np
from pandas import DataFrame, Series
from unfqt.strat_utils import valuewhen
from detecta import detect_peaks


def find_bullish_divergences_at_low_peaks(is_low_peak: Series,
                                          price: Series,
                                          osc: Series,
                                          lookback: int = 1):

    # bullish regular divergence (price LL vs osc HL)
    osc_higher_low = osc > valuewhen(is_low_peak, osc, lookback)
    price_lower_low = price <= valuewhen(is_low_peak, price, lookback)

    found_regular_divergence_mask = osc_higher_low & \
        price_lower_low & is_low_peak

    # hidden divergence (price HL vs osc LL)
    osc_lower_low = osc < valuewhen(is_low_peak, osc, lookback)
    price_higher_low = price > valuewhen(is_low_peak, price, lookback)

    found_hidden_divergence_mask = osc_lower_low & \
        price_higher_low & is_low_peak

    # replace True with 1.0 and False with NaN
    # [:-1]: is needed, because is_low_peak is
    # using duplicated last candle for detect_peaks
    # TODO: refactor `[:-1]` to level above, so that logic isn't split.
    regular = np.where(found_regular_divergence_mask[:-1], 1, np.NaN)
    hidden = np.where(found_hidden_divergence_mask[:-1], 1, np.NaN)
    return regular, hidden


def detect_bullish_divergences_with_lookback(
        dataframe,
        price_label='low',
        osc_label='rsi',
        *,
        label_regular="bullish_regular_divergence",
        label_hidden="bullish_hidden_divergence",
        lookback: int = 1):
    dataframe = dataframe.copy()
    regular = Series(data=None, index=dataframe.index, dtype=np.float64)
    hidden = Series(data=None, index=dataframe.index, dtype=np.float64)
    # we need to duplicate last entries of series
    # to enable detect_peak to work with last candle
    price = Series([*dataframe[price_label], dataframe[price_label].iat[-1]])
    osc = Series([*dataframe[osc_label], dataframe[osc_label].iat[-1]])
    # find peaks and re-append index
    is_low_peak = price.iloc[detect_peaks(
        price, valley=True, show=False, mpd=2, threshold=1)] != np.NaN

    for lookback_i in range(0, lookback + 1):
        reg, hid = find_bullish_divergences_at_low_peaks(is_low_peak,
                                                         price,
                                                         osc,
                                                         lookback_i)
        regular = Series(regular).combine_first(Series(reg))
        hidden = Series(hidden).combine_first(Series(hid))

    dataframe[label_regular] = regular
    dataframe[label_hidden] = hidden
    return regular, hidden, dataframe


def combine_bullish_divergences(dataframe,
                                price_label: str = 'low',
                                oscillators: List[str] = ['rsi'],
                                lookback: int = 1):
    regular = Series(data=0, index=dataframe.index, dtype=np.float64)
    hidden = Series(data=0, index=dataframe.index, dtype=np.float64)
    for osc in oscillators:
        reg, hid, _ = detect_bullish_divergences_with_lookback(
            dataframe, price_label, osc, lookback=lookback)
        regular = regular + reg.fillna(0)
        hidden = hidden + hid.fillna(0)
    return regular, hidden


def find_bearish_divergences_at_low_peaks(is_high_peak: Series,
                                          price: Series,
                                          osc: Series,
                                          lookback: int = 1):

    # bearish regular divergence (price LL vs osc HL)
    osc_lower_high = osc < valuewhen(is_high_peak, osc, lookback)
    price_higher_high = price >= valuewhen(is_high_peak, price, lookback)

    found_regular_divergence_mask = osc_lower_high & \
        price_higher_high & is_high_peak

    # hidden divergence (price HL vs osc LL)
    osc_higher_high = osc > valuewhen(is_high_peak, osc, lookback)
    price_lower_high = price < valuewhen(is_high_peak, price, lookback)

    found_hidden_divergence_mask = osc_higher_high & \
        price_lower_high & is_high_peak

    # replace True with 1.0 and False with NaN
    # NOTE: [:-1]: is needed, because is_low_peak is
    # using duplicated last candle for detect_peaks
    # TODO: refactor `[:-1]` to level above, so that logic isn't split.
    regular = np.where(found_regular_divergence_mask[:-1], 1, np.NaN)
    hidden = np.where(found_hidden_divergence_mask[:-1], 1, np.NaN)
    return regular, hidden


def detect_bearish_divergences_with_lookback(
        dataframe,
        price_label='high',
        osc_label='rsi',
        *,
        label_regular="bearish_regular_divergence",
        label_hidden="bearish_hidden_divergence",
        lookback: int = 1):
    dataframe = dataframe.copy()
    regular = Series(data=None, index=dataframe.index, dtype=np.float64)
    hidden = Series(data=None, index=dataframe.index, dtype=np.float64)
    # we need to duplicate last entries of series
    # to enable detect_peak to work with last candle
    price = Series([*dataframe[price_label], dataframe[price_label].iat[-1]])
    osc = Series([*dataframe[osc_label], dataframe[osc_label].iat[-1]])
    # find peaks and re-append index
    is_high_peak = price.iloc[detect_peaks(
        price, valley=False, show=False, mpd=2, threshold=1)] != np.NaN

    for lookback_i in range(0, lookback + 1):
        reg, hid = find_bearish_divergences_at_low_peaks(is_high_peak,
                                                         price,
                                                         osc,
                                                         lookback_i)
        regular = Series(regular).combine_first(Series(reg))
        hidden = Series(hidden).combine_first(Series(hid))

    dataframe[label_regular] = regular
    dataframe[label_hidden] = hidden
    return regular, hidden, dataframe


def combine_bearish_divergences(dataframe,
                                price_label: str = 'high',
                                oscillators: List[str] = ['rsi'],
                                lookback: int = 1):
    regular = Series(data=0, index=dataframe.index, dtype=np.float64)
    hidden = Series(data=0, index=dataframe.index, dtype=np.float64)
    for osc in oscillators:
        reg, hid, _ = detect_bearish_divergences_with_lookback(
            dataframe, price_label, osc, lookback=lookback)
        # regular = regular + reg.fillna(0)
        regular = regular.add(reg.head(len(regular)).fillna(0).values)
        # hidden = hidden + hid.fillna(0)
        hidden = hidden.add(hid.head(len(hidden)).fillna(0).values)
    return regular, hidden



import functools
from typing import Union
import numpy as np
import pandas as pd
from pandas import Series, DataFrame
import pandas
from itertools import compress
from numbers import Number
from typing import Sequence, Optional, Union, Callable


def angle(a, b):
    """Convert angle between points `a` and `b` to an angle in degrees, relative to x axis."""
    deg = np.rad2deg(np.arctan2(b[1] - a[1], b[0] - a[0]))
    deg = (90 - abs(deg)) * (+1 if deg > 0 else -1)
    return deg


def pivot(s: Series, cmp, left=1, right=1):
    conditions = []
    for i in range(left):
        conditions.append(cmp(s, s.shift(+i + 1)))
    for i in range(right):
        conditions.append(cmp(s, s.shift(-i - 1)))
    return functools.reduce(lambda a, b: a & b, conditions)


def pivotlow(s: Series, left=1, right=1):
    return pivot(s, np.less, left, right)


def pivothigh(s: Series, left=1, right=1):
    return pivot(s, np.greater, left, right)


def repeat_condition(self, series: pandas.Series, length: int):
    '''
    HOWTO:
    dataframe.loc[
        (
            (dataframe['go_long'] > 1) &
            repeat_condition(dataframe['minus_di_5m'] < 1.2 * dataframe['minus_di_5m'].shift(1), 5)
        ),
        'buy'] = 0
    '''
    conditions = []
    for i in range(length):
        conditions.append(series.shift(i))
    return functools.reduce(lambda a, b: a & b, conditions)


def line2arr(line, size=-1):
    '''
    Creates an numpy array from a backtrader line

    This method wraps the lines array in numpy. This can
    be used for conditions.
    '''
    if size <= 0:
        return np.array(line.array)
    else:
        return np.array(line.get(size=size))


def na(val):
    '''
    RETURNS
    true if x is not a valid number (x is NaN), otherwise false.
    '''
    return val != val


def nz(x, y=None):
    '''
    RETURNS
    Two args version: returns x if it's a valid (not NaN) number, otherwise y
    One arg version: returns x if it's a valid (not NaN) number, otherwise 0
    ARGUMENTS
    x (val) Series of values to process.
    y (float) Value that will be inserted instead of all NaN values in x series.
    '''
    if isinstance(x, np.generic):
        return x.fillna(y or 0)
    if x != x:
        if y is not None:
            return y
        return 0
    return x


def barssince(condition: Sequence[bool], default=np.inf) -> int:
    """
    Return the number of bars since `condition` sequence was last `True`,
    or if never, return `default`.
        >>> barssince(self.data.Close > self.data.Open)
        3
    """
    return next(compress(range(len(condition)), reversed(condition)), default)


def barssince_while(condition, occurrence=0):
    '''
    Impl of barssince

    RETURNS
    Number of bars since condition was true.
    REMARKS
    If the condition has never been met prior to the current bar, the function returns na.
    '''
    cond_len = len(condition)
    occ = 0
    since = 0
    res = float('nan')
    while cond_len - (since+1) >= 0:
        cond = condition[cond_len-(since+1)]
        # check for nan cond != cond == True when nan
        if cond and not cond != cond:
            if occ == occurrence:
                res = since
                break
            occ += 1
        since += 1
    return res


def valuewhen(condition, source, occurrence=0):
    # NOTE: checked, not .shift(-occurrence) is correct
    return source \
        .reindex(condition[condition].index) \
        .shift(occurrence) \
        .reindex(source.index) \
        .ffill()


def valuewhen_iter(condition, source, occurrence=0):
    '''
    Impl of valuewhen
    + added occurrence

    RETURNS
    Source value when condition was true
    '''
    res = float('nan')
    since = barssince(condition, occurrence)
    if since is not None:
        res = source[-(since+1)]
    return res


def calc_streaks(series: pd.Series):

    # logic tables
    geq = series >= series.shift(1)  # True if rising
    eq = series == series.shift(1)  # True if equal
    logic_table = pd.concat([geq, eq], axis=1)

    streaks = [0]  # holds the streak duration, starts with 0

    for row in logic_table.iloc[1:].itertuples():  # iterate through logic table
        if row[2]:  # same value as before
            streaks.append(0)
            continue
        last_value = streaks[-1]
        if row[1]:  # higher value than before
            streaks.append(last_value + 1 if last_value >=
                           0 else 1)  # increase or reset to +1
        else:  # lower value than before
            streaks.append(last_value - 1 if last_value <
                           0 else -1)  # decrease or reset to -1

    return streaks


def cross_up(a: Series, b: Union[Series, int, float]):
    b_prev = b.shift(1) if isinstance(b, Series) else b
    return (a.shift(1) <= b_prev) & (a > b)


def cross_down(a: Series, b: Union[Series, int, float]):
    b_prev = b.shift(1) if isinstance(b, Series) else b
    return (a.shift(1) >= b_prev) & (a < b)


def cross_any(a: Series, b: Union[Series, int, float]):
    return cross_up(a, b) | cross_down(a, b)


def going_up(series: Series, period: int):
    conditions = []
    for i in range(period):
        conditions.append(series.shift(i + 1) < series.shift(i))
    return functools.reduce(lambda a, b: a & b, conditions)


def going_down(series: Series, period: int):
    conditions = []
    for i in range(period):
        conditions.append(series.shift(i + 1) > series.shift(i))
    return functools.reduce(lambda a, b: a & b, conditions)


def typical_price(dataframe: DataFrame):
    return (dataframe['high'] + dataframe['low'] + dataframe['close']) / 3


def repeat_condition(condition: Series, length: int, op: str = 'and', step: int = 1):
    """
    Repeat condition `length` times shifting it by `step`.
    :param condition:
    :param length:
    :param op: 'or' or 'and'
    :param step:
    :return:
    """
    conditions = []
    for i in range(length):
        conditions.append(condition.shift(i * step))
    if op == 'or':
        def reducer(a, b):
            return a | b
    elif op == 'and':
        def reducer(a, b):
            return a & b
    else:
        raise Exception('op should be "or" or "and"')
    return functools.reduce(reducer, conditions)


def value_when(condition, source, occurrence):
    return source \
        .reindex(condition[condition].index) \
        .shift(-occurrence) \
        .reindex(source.index) \
        .ffill()


def heikin_ashi(dataframe: DataFrame):
    hk = DataFrame()
    hk['close'] = (dataframe['open'] + dataframe['close'] +
                   dataframe['high'] + dataframe['low']) / 4
    hk['open'] = (dataframe['open'].shift(1) + dataframe['close'].shift(1)) / 2
    hk['low'] = dataframe['low']
    hk['low'] = hk[['low', 'open', 'close']].min(axis=1)
    hk['high'] = dataframe['high']
    hk['high'] = hk[['high', 'open', 'close']].max(axis=1)
    return hk


def ssl(high, low, close, length=10):
    df = DataFrame()
    df['sma_high'] = ta.sma(high, length=length)
    df['sma_low'] = ta.sma(low, length=length)
    df['hlv'] = np.where(close > df['sma_high'], 1,
                         np.where(close < df['sma_low'], -1, np.NAN))
    df['hlv'] = df['hlv'].ffill()
    df['ssl_down'] = np.where(df['hlv'] < 0, df['sma_high'], df['sma_low'])
    df['ssl_up'] = np.where(df['hlv'] < 0, df['sma_low'], df['sma_high'])
    return df['ssl_down'], df['ssl_up']


def bov(df: DataFrame):
    """
    Balance of volume indicator by @rk. Splits entire volume into buying and selling volume.
    :param df: dataframe.
    :return: (buying_volume, selling_volume)
    """
    v: Series = df['volume']
    power = ta.bop(df['open'], df['high'], df['low'], df['close'])
    # bop range is [-1, +1], convert it to [0, +1]
    power = 0.5 + power / 2
    vol_up = v * power              # volume in the direction of the candle
    # volume against the direction of the candle
    vol_dn = v * (1 - power)
    return vol_up, vol_dn


def vres(df: DataFrame):
    """
    Volume-derived price movement resistance indicator by @rk. Measures how much volume
    is required to move price. Indicator value range is undefined. Values should be
    interpreted by comparing them to one another. Value fluctuates from negative to
    positive, sign indicating candle direction. Values closer to 0 mean reduced
    resistance while values further from 0 mean increased resistance.

    Consider two candles, red and green, of equal size. vres_red.abs() > vres_green.abs()
    signals that downwards resistance is greater and it is easier for price to move
    up, therefore it is a bullish sign.

    :param df:
    :return:
    """
    v = df['volume']
    h = df['high'] - df['low']
    power = ta.bop(df['open'], df['high'], df['low'], df['close'])
    # bop range is [-1, +1], convert it to [0, +1], < 0.5 - sellers in control, > 0.5 - buyers in control
    power = 0.5 + power / 2
    # Convert power to % of control in direction of the candle.
    power = np.where(df['open'] < df['close'], 1 - power, power)
    return v * power / h


def vwbop(df: DataFrame, length=14, kind='sma'):
    """
    Volume-weighted balance of power indicator by @rk.
    :param df:
    :return:
    """
    bop = ta.bop(df['open'], df['high'], df['low'], df['close'])
    volume = df['volume'] * bop.abs()
    pv = bop * volume
    if length > 1:
        return ta.ma(kind, close=pv, length=length) / ta.ma(kind, close=volume, length=length)
    return pv


def swbop(df: DataFrame, length=14, kind='sma'):
    """
    Candle-size-weighted balance of power indicator by @rk.
    :param df:
    :param length:
    :param kind:
    :return:
    """
    bop = ta.bop(df['open'], df['high'], df['low'], df['close'])
    size = df['high'] - df['low']
    pv = bop * size
    return ta.ma(kind, close=pv, length=length) / ta.ma(kind, close=size, length=length)


def vwtp(df: DataFrame, length=14, kind='sma'):
    """
    Volume-weighted typical price.
    :param df:
    :param length:
    :param kind:
    :return:
    """
    tprice = (df['high'] + df['low'] + df['close']) / 3
    volume = df['volume']
    pv = tprice * volume
    return ta.ma(kind, close=pv, length=length) / ta.ma(kind, close=volume, length=length)



"""
Solipsis Custom Indicators and Maths
Some indicators are direct copies from freqtrade/technical but I did not want to rely on that repository
to be stable and I needed to make some fixes/changes to fibonacci retracements and PMAX to suit my needs.

extended by @TheJoeSchr 
"""
from typing import List, Union
from pandas import Series
import pandas as pd
import numpy as np
import talib.abstract as ta
import freqtrade.vendor.qtpylib.indicators as qtpylib

from pandas import DataFrame, Series
from pandas_ta.utils import get_drift, get_offset, verify_series, zero
from pandas_ta.volatility import atr
from pandas_ta.overlap import ma
# from pandas_ta import adx


def find_higher_tf_divergences(dataframe,
                               lower_tf_postfix='_btc_5m',
                               higher_tf_postfix='_btc_1h'):
    dataframe['low_peaks'+lower_tf_postfix] = np.NaN
    dataframe['high_peaks'+lower_tf_postfix] = np.NaN
    dataframe.loc[dataframe.index.isin(detect_peaks(
        dataframe['high'+lower_tf_postfix][:-1], valley=False, show=False)), 'high_peaks'+lower_tf_postfix] = dataframe['high'+lower_tf_postfix]
    dataframe.loc[dataframe.index.isin(detect_peaks(
        dataframe['low'+lower_tf_postfix][:-1], valley=True, show=False)), 'low_peaks'+lower_tf_postfix] = dataframe['low'+lower_tf_postfix]

    dataframe['peaks_osc'+lower_tf_postfix] = np.where(
        dataframe['low_peaks'+lower_tf_postfix] == dataframe['low'+lower_tf_postfix], dataframe['osc'+lower_tf_postfix], np.NaN)
    dataframe['peaks_osc'+higher_tf_postfix] = np.where(
        dataframe['low_peaks'+higher_tf_postfix].shift(-12) == dataframe['low'+lower_tf_postfix], dataframe['osc'+lower_tf_postfix], np.NaN)
    # index.isin: brings in latest 5m low peak
    dataframe['peaks_osc'+higher_tf_postfix+'_shifted'] = dataframe.loc[(dataframe['low_peaks'+higher_tf_postfix].shift(-12) == dataframe['low'+lower_tf_postfix]) | (
        dataframe.index.isin(dataframe.loc[dataframe['low_peaks'+lower_tf_postfix].notna()][-1::].index.values)), 'osc'+lower_tf_postfix].shift()
    # index.isin: brings in latest 5m low peak
    dataframe['peaks_low'+higher_tf_postfix+'_shifted'] = dataframe.loc[(dataframe['low_peaks'+higher_tf_postfix].shift(-12) == dataframe['low'+lower_tf_postfix]) | (
        dataframe.index.isin(dataframe.loc[dataframe['low_peaks'+lower_tf_postfix].notna()][-1::].index.values)), 'low'+lower_tf_postfix].shift()

    dataframe['found_divergence_hidden'+lower_tf_postfix] = 0
    dataframe['found_divergence_regular'+lower_tf_postfix] = 0

    dataframe.loc[(dataframe['peaks_low'+higher_tf_postfix+'_shifted'] > dataframe['low'+lower_tf_postfix]) & (
        dataframe['peaks_osc'+higher_tf_postfix+'_shifted'] <= dataframe['peaks_osc'+lower_tf_postfix]), 'found_divergence_regular'+lower_tf_postfix] = 1

    dataframe.loc[(dataframe['peaks_low'+higher_tf_postfix+'_shifted'] < dataframe['low'+lower_tf_postfix]) & (
        dataframe['peaks_osc'+higher_tf_postfix+'_shifted'] >= dataframe['peaks_osc'+lower_tf_postfix]), 'found_divergence_hidden'+lower_tf_postfix] = 1

    dataframe.loc[(dataframe['found_divergence_hidden'+lower_tf_postfix].shift() == 1) | (
        dataframe['found_divergence_regular'+lower_tf_postfix].shift() == 1), 'found_divergence_marker'+lower_tf_postfix] = dataframe['low'+lower_tf_postfix]

    # makes rolling().sum() easier than with NaN
    dataframe['found_divergence_marker' +
              lower_tf_postfix].fillna(0, inplace=True)
    return dataframe


def vwap_fast(dataframe: DataFrame):
    """
    by @rk
    """
    split_indices = list(dataframe.loc[
        (
            (dataframe['date'].dt.second == 0)
            &
            (dataframe['date'].dt.minute == 0)
            &
            (dataframe['date'].dt.hour == 0)
        )].index)
    split_indices.insert(0, 0)
    split_indices.append(len(dataframe))
    vwap_slices = []
    for i in range(1, len(split_indices)):
        start_idx = split_indices[i - 1]
        end_idx = split_indices[i]
        slice = dataframe[start_idx:end_idx]
        hlc3 = (slice['high'] + slice['low'] + slice['close']) / 3
        wp = hlc3 * slice['volume']
        vwap = wp.cumsum() / slice['volume'].cumsum()
        vwap_slices.append(vwap)
    vwap = pd.concat(vwap_slices)
    return vwap


def williams_fractal(dataframe):
    """
    DANGEROUS: needs to be used with .shift(2) because of lookahead bias
    add 'fractals_high' and 'fractals_low' to dataframe
    """
    dataframe.loc[(dataframe['high'] > dataframe['high'].shift(1)) &
                  (dataframe['high'] > dataframe['high'].shift(2)) &
                  (dataframe['high'] > dataframe['high'].shift(-1)) &
                  (dataframe['high'] > dataframe['high'].shift(-2)),
                  'fractal_high'
                  ] = 1
    dataframe.loc[((dataframe['low'] < dataframe['low'].shift(1)) &
                   (dataframe['low'] < dataframe['low'].shift(2)) &
                   (dataframe['low'] < dataframe['low'].shift(-1)) &
                   (dataframe['low'] < dataframe['low'].shift(-2))),
                  'fractal_low'] = 1
    return dataframe


"""
Misc. Helper Functions
"""


def same_length(bigger, shorter):
    return np.concatenate((np.full((bigger.shape[0] - shorter.shape[0]), np.nan), shorter))


"""
Maths
"""


def linear_growth(start: float, end: float, start_time: int, end_time: int, trade_time: int) -> float:
    """
    Simple linear growth function. Grows from start to end after end_time minutes (starts after start_time minutes)
    """
    time = max(0, trade_time - start_time)
    rate = (end - start) / (end_time - start_time)

    return min(end, start + (rate * time))


"""
TA Indicators
"""


def fib_ret(df, lookback=72, field='close') -> DataFrame:
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/indicators.py#L735
    Modifed to have a specific lookback period to prevent lookahead bias.
    """
    thresholds = [1.0, 0.786, 0.618, 0.5, 0.382, 0.236, 0.0]

    window_min = df[field].rolling(lookback).min()
    window_max = df[field].rolling(lookback).max()
    # I have no idea what this is for, maaybe graphing or jupyter?
    fib_levels = [window_min + t * (window_max - window_min)
                  for t in thresholds]

    data = (df[field] - window_min) / (window_max - window_min)

    # this return is broken and I don't know how to fix it.
    # return data.apply(lambda x: max(t for t in thresholds if x >= t))
    return data


def RMI(dataframe, *, length=20, mom=5):
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/indicators.py#L912
    """
    df = dataframe.copy()

    df['maxup'] = (df['close'] - df['close'].shift(mom)).clip(lower=0)
    df['maxdown'] = (df['close'].shift(mom) - df['close']).clip(lower=0)

    df.fillna(0, inplace=True)

    df["emaInc"] = ta.EMA(df, price='maxup', timeperiod=length)
    df["emaDec"] = ta.EMA(df, price='maxdown', timeperiod=length)

    df['RMI'] = np.where(df['emaDec'] == 0, 0, 100 - 100 /
                         (1 + df["emaInc"] / df["emaDec"]))

    return df["RMI"]


def VIDYA(dataframe, length=9, select=True):
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/indicators.py#L936
    """
    df = dataframe.copy()

    alpha = 2 / (length + 1)
    df['momm'] = df['close'].diff()
    df['m1'] = np.where(df['momm'] >= 0, df['momm'], 0.0)
    df['m2'] = np.where(df['momm'] >= 0, 0.0, -df['momm'])

    df['sm1'] = df['m1'].rolling(length).sum()
    df['sm2'] = df['m2'].rolling(length).sum()

    df['chandeMO'] = 100 * (df['sm1'] - df['sm2']) / (df['sm1'] + df['sm2'])
    if select:
        df['k'] = abs(df['chandeMO']) / 100
    else:
        df['k'] = df['close'].rolling(length).std()
    df.fillna(0.0, inplace=True)

    df['VIDYA'] = 0.0
    for i in range(length, len(df)):

        df['VIDYA'].iat[i] = alpha * df['k'].iat[i] * df['close'].iat[i] + \
            (1 - alpha * df['k'].iat[i]) * df['VIDYA'].iat[i - 1]

    return df['VIDYA']


def vwma(df, window):
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/overlap_studies.py#L75
    """
    return (df['close'] * df['volume']).rolling(window).sum() / df.volume.rolling(window).sum()


def zema(dataframe, period, field='close'):
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/overlap_studies.py#L79
    Modified slightly to use ta.EMA instead of technical ema
    """
    df = dataframe.copy()

    df['ema1'] = ta.EMA(df[field], timeperiod=period)
    df['ema2'] = ta.EMA(df['ema1'], timeperiod=period)
    df['d'] = df['ema1'] - df['ema2']
    df['zema'] = df['ema1'] + df['d']

    return df['zema']


def PMAX(dataframe, period=10, multiplier=3, length=12, MAtype=1, src=1):
    """
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/indicators.py#L1059
    Modified to return different series instead of a modified dataframe as well as not use unique
    field names based on the input parameters.
    returns: atr, ma, pmax_value, pmax_location
    src==1 --> close
    src==2 --> hl2
    src==3 --> ohlc4
    MAtype==1 --> EMA
    MAtype==2 --> DEMA
    MAtype==3 --> T3
    MAtype==4 --> SMA
    MAtype==5 --> VIDYA
    MAtype==6 --> TEMA
    MAtype==7 --> WMA
    MAtype==8 --> VWMA
    MAtype==9 --> zema
    """
    df = dataframe.copy()
    mavalue = 'ma'
    atr = 'atr'
    pm = 'pm'
    pmx = 'pmx'

    df[atr] = ta.ATR(df, timeperiod=period)

    if src == 1:
        masrc = df["close"]
    elif src == 2:
        masrc = (df["high"] + df["low"]) / 2
    elif src == 3:
        masrc = (df["high"] + df["low"] + df["close"] + df["open"]) / 4

    if MAtype == 1:
        df[mavalue] = ta.EMA(masrc, timeperiod=length)
    elif MAtype == 2:
        df[mavalue] = ta.DEMA(masrc, timeperiod=length)
    elif MAtype == 3:
        df[mavalue] = ta.T3(masrc, timeperiod=length)
    elif MAtype == 4:
        df[mavalue] = ta.SMA(masrc, timeperiod=length)
    elif MAtype == 5:
        df[mavalue] = VIDYA(df, length=length)
    elif MAtype == 6:
        df[mavalue] = ta.TEMA(masrc, timeperiod=length)
    elif MAtype == 7:
        df[mavalue] = ta.WMA(df, timeperiod=length)
    elif MAtype == 8:
        df[mavalue] = vwma(df, length)
    elif MAtype == 9:
        df[mavalue] = zema(df, period=length)

    # Compute basic upper and lower bands
    df['basic_ub'] = df[mavalue] + (multiplier * df[atr])
    df['basic_lb'] = df[mavalue] - (multiplier * df[atr])
    # Compute final upper and lower bands
    df['final_ub'] = 0.00
    df['final_lb'] = 0.00
    for i in range(period, len(df)):
        df['final_ub'].iat[i] = df['basic_ub'].iat[i] if (
            df['basic_ub'].iat[i] < df['final_ub'].iat[i - 1]
            or df[mavalue].iat[i - 1] > df['final_ub'].iat[i - 1]) else df['final_ub'].iat[i - 1]
        df['final_lb'].iat[i] = df['basic_lb'].iat[i] if (
            df['basic_lb'].iat[i] > df['final_lb'].iat[i - 1]
            or df[mavalue].iat[i - 1] < df['final_lb'].iat[i - 1]) else df['final_lb'].iat[i - 1]

    # Set the Pmax value
    df[pm] = 0.00
    for i in range(period, len(df)):
        df[pm].iat[i] = (
            df['final_ub'].iat[i] if (df[pm].iat[i - 1] == df['final_ub'].iat[i - 1]
                                      and df[mavalue].iat[i] <= df['final_ub'].iat[i])
            else df['final_lb'].iat[i] if (
                df[pm].iat[i - 1] == df['final_ub'].iat[i - 1]
                and df[mavalue].iat[i] > df['final_ub'].iat[i]) else df['final_lb'].iat[i]
            if (df[pm].iat[i - 1] == df['final_lb'].iat[i - 1]
                and df[mavalue].iat[i] >= df['final_lb'].iat[i]) else df['final_ub'].iat[i]
            if (df[pm].iat[i - 1] == df['final_lb'].iat[i - 1]
                and df[mavalue].iat[i] < df['final_lb'].iat[i]) else 0.00)

    # Mark the trend direction up/down
    df[pmx] = np.where((df[pm] > 0.00), np.where(
        (df[mavalue] < df[pm]), 'down', 'up'), np.NaN)
    # Remove basic and final bands from the columns
    df.drop(['basic_ub', 'basic_lb', 'final_ub',
            'final_lb'], inplace=True, axis=1)

    df.fillna(0, inplace=True)

    return df[atr], df[mavalue], df[pm], df[pmx]


def macross(dataframe: DataFrame, fast: int = 20, slow: int = 50) -> Series:
    """
    Moving Average Cross
    Port of: https://www.tradingview.com/script/PcWAuplI-Moving-Average-Cross/
    """
    df = dataframe.copy()

    # Fast MAs
    upper_fast = ta.EMA(df['high'], timeperiod=fast)
    lower_fast = ta.EMA(df['low'], timeperiod=fast)

    # Slow MAs
    upper_slow = ta.EMA(df['high'], timeperiod=slow)
    lower_slow = ta.EMA(df['low'], timeperiod=slow)

    # Crosses
    crosses_lf_us = qtpylib.crossed_above(
        lower_fast, upper_slow) | qtpylib.crossed_below(lower_fast, upper_slow)
    crosses_uf_ls = qtpylib.crossed_above(
        upper_fast, lower_slow) | qtpylib.crossed_below(upper_fast, lower_slow)

    dir_1 = np.where(crosses_lf_us, 1, np.nan)
    dir_2 = np.where(crosses_uf_ls, -1, np.nan)

    dir = np.where(dir_1 == 1, dir_1, np.nan)
    dir = np.where(dir_2 == -1, dir_2, dir_1)

    res = Series(dir).fillna(method='ffill').to_numpy()

    return res


def mastreak(dataframe: DataFrame, period: int = 4, field='close') -> Series:
    """
    MA Streak
    Port of: https://www.tradingview.com/script/Yq1z7cIv-MA-Streak-Can-Show-When-a-Run-Is-Getting-Long-in-the-Tooth/
    """
    df = dataframe.copy()

    avgval = zema(df, period, field)

    arr = np.diff(avgval)
    pos = np.clip(arr, 0, 1).astype(bool).cumsum()
    neg = np.clip(arr, -1, 0).astype(bool).cumsum()
    streak = np.where(arr >= 0, pos - np.maximum.accumulate(np.where(arr <= 0, pos, 0)),
                      -neg + np.maximum.accumulate(np.where(arr >= 0, neg, 0)))

    res = same_length(df['close'], streak)

    return res


def pcc(dataframe: DataFrame, period: int = 20, mult: int = 2):
    """
    Percent Change Channel
    PCC is like KC unless it uses percentage changes in price to set channel distance.
    https://www.tradingview.com/script/6wwAWXA1-MA-Streak-Change-Channel/
    """
    df = dataframe.copy()

    df['previous_close'] = df['close'].shift()

    df['close_change'] = (df['close'] - df['previous_close']
                          ) / df['previous_close'] * 100
    df['high_change'] = (df['high'] - df['close']) / df['close'] * 100
    df['low_change'] = (df['low'] - df['close']) / df['close'] * 100

    df['delta'] = df['high_change'] - df['low_change']

    mid = zema(df, period, 'close_change')
    rangema = zema(df, period, 'delta')

    upper = mid + rangema * mult
    lower = mid - rangema * mult

    return upper, rangema, lower


def SSLChannels(dataframe, length=10, mode='sma'):
    """
    Source: https://www.tradingview.com/script/xzIoaIJC-SSL-channel/
    Source: https://github.com/freqtrade/technical/blob/master/technical/indicators/indicators.py#L1025
    Author: xmatthias
    Pinescript Author: ErwinBeckers
    SSL Channels.
    Average over highs and lows form a channel - lines "flip" when close crosses either of the 2 lines.
    Trading ideas:
        * Channel cross
        * as confirmation based on up > down for long
    Usage:
        dataframe['sslDown'], dataframe['sslUp'] = SSLChannels(dataframe, 10)
    """
    if mode not in ('sma'):
        raise ValueError(f"Mode {mode} not supported yet")

    df = dataframe.copy()

    if mode == 'sma':
        df['smaHigh'] = df['high'].rolling(length).mean()
        df['smaLow'] = df['low'].rolling(length).mean()

    df['hlv'] = np.where(df['close'] > df['smaHigh'], 1,
                         np.where(df['close'] < df['smaLow'], -1, np.NAN))
    df['hlv'] = df['hlv'].ffill()

    df['sslDown'] = np.where(df['hlv'] < 0, df['smaHigh'], df['smaLow'])
    df['sslUp'] = np.where(df['hlv'] < 0, df['smaLow'], df['smaHigh'])

    return df['sslDown'], df['sslUp']

# From CombinedBinHAAndCluc


def bollinger_bands(stock_price, window_size, num_of_std):
    rolling_mean = stock_price.rolling(window=window_size).mean()
    rolling_std = stock_price.rolling(window=window_size).std()
    lower_band = rolling_mean - (rolling_std * num_of_std)
    return np.nan_to_num(rolling_mean), np.nan_to_num(lower_band)


"""
Market Cipher, from @drafty in Freqtrade discord
"""


def market_cipher(self, dataframe) -> DataFrame:
    #dataframe['volume_rolling'] = dataframe['volume'].shift(14).rolling(14).mean()
    #
    dataframe['ap'] = (dataframe['high'] +
                       dataframe['low'] + dataframe['close']) / 3
    dataframe['esa'] = ta.EMA(dataframe['ap'], self.n1)
    dataframe['d'] = ta.EMA(
        (dataframe['ap'] - dataframe['esa']).abs(), self.n1)
    dataframe['ci'] = (dataframe['ap'] - dataframe['esa']) / \
        (0.015 * dataframe['d'])
    dataframe['tci'] = ta.EMA(dataframe['ci'], self.n2)

    dataframe['wt1'] = dataframe['tci']
    dataframe['wt2'] = ta.SMA(dataframe['wt1'], 4)
    dataframe['wt1-wt2'] = dataframe['wt1'] - dataframe['wt2']

    dataframe['crossed_above'] = qtpylib.crossed_above(
        dataframe['wt2'], dataframe['wt1'])
    dataframe['crossed_below'] = qtpylib.crossed_below(
        dataframe['wt2'], dataframe['wt1'])
    #dataframe['slope_gd'] = ta.LINEARREG_ANGLE(dataframe['crossed_above'] * dataframe['wt2'], 10)

    return dataframe


def EWO(dataframe, sma_length=5, sma2_length=35):
    df = dataframe.copy()
    sma1 = ta.SMA(df, timeperiod=sma_length)
    sma2 = ta.SMA(df, timeperiod=sma2_length)
    smadif = (sma1 - sma2) / df['close'] * 100
    return smadif


def adx(high, low, close, length=None, scalar=None, mamode=None, drift=None, offset=None, lensig=None, **kwargs):
    """Indicator: ADX"""
    # Validate Arguments
    length = length if length and length > 0 else 14
    lensig = lensig if lensig and lensig > 0 else length
    mamode = mamode if isinstance(mamode, str) else "rma"
    scalar = float(scalar) if scalar else 100
    high = verify_series(high)
    low = verify_series(low)
    close = verify_series(close)
    drift = get_drift(drift)
    offset = get_offset(offset)

    if high is None or low is None or close is None:
        return

    # Calculate Result
    atr_ = atr(high=high, low=low, close=close, length=length)

    up = high - high.shift(drift)  # high.diff(drift)
    dn = low.shift(drift) - low    # low.diff(-drift).shift(drift)

    pos = ((up > dn) & (up > 0)) * up
    neg = ((dn > up) & (dn > 0)) * dn

    pos = pos.apply(zero)
    neg = neg.apply(zero)

    k = scalar / atr_
    dmp = k * ma(mamode, pos, length=length)
    dmn = k * ma(mamode, neg, length=length)

    dx = scalar * (dmp - dmn).abs() / (dmp + dmn)
    adx = ma(mamode, dx, length=lensig)

    # Offset
    if offset != 0:
        dmp = dmp.shift(offset)
        dmn = dmn.shift(offset)
        adx = adx.shift(offset)

    # Handle fills
    if "fillna" in kwargs:
        adx.fillna(kwargs["fillna"], inplace=True)
        dmp.fillna(kwargs["fillna"], inplace=True)
        dmn.fillna(kwargs["fillna"], inplace=True)
    if "fill_method" in kwargs:
        adx.fillna(method=kwargs["fill_method"], inplace=True)
        dmp.fillna(method=kwargs["fill_method"], inplace=True)
        dmn.fillna(method=kwargs["fill_method"], inplace=True)

    # Name and Categorize it
    adx.name = f"ADX_{lensig}"
    dmp.name = f"DMP_{length}"
    dmn.name = f"DMN_{length}"

    adx.category = dmp.category = dmn.category = "trend"

    # Prepare DataFrame to return
    data = {adx.name: adx, dmp.name: dmp, dmn.name: dmn}
    adxdf = DataFrame(data)
    adxdf.name = f"ADX_{lensig}"
    adxdf.category = "trend"

    return adxdf



import pandas as pd

from freqtrade.exchange import timeframe_to_minutes, timeframe_to_prev_date
from freqtrade.strategy import IStrategy
from datetime import datetime, timedelta
from freqtrade.persistence import Trade

from functools import wraps

# anti-future bias wrapper
def analyze_df_in_batches(f, window=500):
    @wraps(f)
    def wrapper(df, *args, **kwargs):
        lines = []
        for i in range(len(df) - window):
            slice = df[i : i + window]
            res = f(slice, *args, **kwargs)
            lines.append(res.iloc[[-1]])
        return pd.concat(lines)

    return wrapper


def get_trade_opened_candle(self, trade: "Trade"):
    """
    search for nearest row of trade.open_date
    """
    trade_candle = find_candle_near_datetime(self, trade.open_date_utc, pair=trade.pair)
    return trade_candle


def get_buy_signal_candle(self, trade: "Trade", timeframe="5m"):
    """
    search for nearest buy candle next to trade.open_date
    """
    buy_candle = find_candle_near_datetime(
        self,
        trade.open_date_utc - timedelta(minutes=timeframe_to_minutes(self.timeframe)),
        pair=trade.pair,
    )
    return buy_candle


def find_candle_near_datetime(self, query_date: datetime, pair: str):
    result = None
    # get dataframe
    dataframe, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
    candle = search_dataframe_index(
        self,
        query_date,
        dataframe,
    )
    result = candle if candle.empty else candle.squeeze()
    return result


def search_dataframe_index(self, query_date: datetime, dataframe):
    df = dataframe[["date"]].set_index("date")
    try:
        date_mask = df.index.unique().get_loc(query_date, method="ffill")
        candle = dataframe.iloc[date_mask]  # use iloc because date_mask :int
    except KeyError:  # trade.open_date may not exist if trade hasn't been opened yet
        candle = pd.DataFrame(index=dataframe.index)
    return candle


def find_candle_datetime_faster(self, query_date: datetime, now: datetime, dataframe):
    """
    fast, but looks into future,
    if dataframe hasn't been sanitzed
    via e.g. using self.dp.get_analyzed_dataframe
    """

    if now and now == query_date:
        candle = dataframe.iloc[-1]
    else:
        candle_date = timeframe_to_prev_date(self.timeframe, query_date)
        candle = dataframe.loc[dataframe.date == candle_date]
    return candle



import pandas as pd

from freqtrade.exchange import timeframe_to_minutes, timeframe_to_prev_date
from freqtrade.strategy import IStrategy
from datetime import datetime
from freqtrade.persistence import Trade


def merge_informative_pair(dataframe: pd.DataFrame, informative: pd.DataFrame,
                           timeframe: str, timeframe_inf: str, ffill: bool = True) -> pd.DataFrame:
    """
    Correctly merge informative samples to the original dataframe, avoiding lookahead bias.

    Since dates are candle open dates, merging a 15m candle that starts at 15:00, and a
    1h candle that starts at 15:00 will result in all candles to know the close at 16:00
    which they should not know.

    Moves the date of the informative pair by 1 time interval forward.
    This way, the 14:00 1h candle is merged to 15:00 15m candle, since the 14:00 1h candle is the
    last candle that's closed at 15:00, 15:15, 15:30 or 15:45.

    Assuming inf_tf = '1d' - then the resulting columns will be:
    date_1d, open_1d, high_1d, low_1d, close_1d, rsi_1d

    :param dataframe: Original dataframe
    :param informative: Informative pair, most likely loaded via dp.get_pair_dataframe
    :param timeframe: Timeframe of the original pair sample.
    :param timeframe_inf: Timeframe of the informative pair sample.
    :param ffill: Forwardfill missing values - optional but usually required
    :return: Merged dataframe
    :raise: ValueError if the secondary timeframe is shorter than the dataframe timeframe
    """

    minutes_inf = timeframe_to_minutes(timeframe_inf)
    minutes = timeframe_to_minutes(timeframe)
    if minutes == minutes_inf:
        # No need to forwardshift if the timeframes are identical
        informative['date_merge'] = informative["date"]
    elif minutes < minutes_inf:
        # Subtract "small" timeframe so merging is not delayed by 1 small candle
        # Detailed explanation in https://github.com/freqtrade/freqtrade/issues/4073
        informative['date_merge'] = (
            informative["date"] + pd.to_timedelta(minutes_inf, 'm') - pd.to_timedelta(minutes, 'm')
        )
    else:
        raise ValueError("Tried to merge a faster timeframe to a slower timeframe."
                         "This would create new rows, and can throw off your regular indicators.")

    # Rename columns to be unique
    informative.columns = [f"{col}_{timeframe_inf}" for col in informative.columns]

    # Combine the 2 dataframes
    # all indicators on the informative sample MUST be calculated before this point
    dataframe = pd.merge(dataframe, informative, left_on='date',
                         right_on=f'date_merge_{timeframe_inf}', how='left')
    dataframe = dataframe.drop(f'date_merge_{timeframe_inf}', axis=1)

    if ffill:
        dataframe = dataframe.ffill()

    return dataframe


def stoploss_from_open(open_relative_stop: float, current_profit: float) -> float:
    """

    Given the current profit, and a desired stop loss value relative to the open price,
    return a stop loss value that is relative to the current price, and which can be
    returned from `custom_stoploss`.

    The requested stop can be positive for a stop above the open price, or negative for
    a stop below the open price. The return value is always >= 0.

    Returns 0 if the resulting stop price would be above the current price.

    :param open_relative_stop: Desired stop loss percentage relative to open price
    :param current_profit: The current profit percentage
    :return: Positive stop loss value relative to current price
    """

    # formula is undefined for current_profit -1, return maximum value
    if current_profit == -1:
        return 1

    stoploss = 1-((1+open_relative_stop)/(1+current_profit))

    # negative stoploss values indicate the requested stop price is higher than the current price
    return max(stoploss, 0.0)


class HelperMixin(IStrategy):
    custom_stoploss_config = {}

    def get_custom_dataframe(self, pair: str):
        dataframe, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
        return dataframe

    def get_trade_candle(self, trade: 'Trade'):
        """
        search for nearest row of trade.open_date
        """
        trade_candle = self.find_candle_datetime(trade.open_date_utc, pair=trade.pair, now=None)
        return trade_candle

    def find_candle_datetime(self, query_date: datetime, pair: str, now: datetime):
        result = None
        dataframe = self.get_custom_dataframe(pair)
        candle = self.find_candle_datetime_safer(query_date, now, dataframe,)
        result = candle if candle.empty else candle.squeeze()
        return result

    def find_candle_datetime_faster(self, query_date: datetime, now: datetime, dataframe):
        if(now and now == query_date):
            candle = dataframe.iloc[-1]
        else:
            candle_date = timeframe_to_prev_date(self.timeframe, query_date)
            candle = dataframe.loc[dataframe.date == candle_date]
        return candle

    def find_candle_datetime_safer(self, query_date: datetime, now: datetime, dataframe):
        df = dataframe[['date']].set_index('date')

        try:
            date_mask = df.index.unique().get_loc(query_date, method='ffill')
            candle = dataframe.iloc[date_mask]  # use iloc because date_mask maybe :int
        except KeyError:  # trade.open_date may not exist yet
            candle = pd.DataFrame(index=dataframe.index)
        return candle

    def __init__(self, config: dict) -> None:
        super().__init__(config)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['unfqt']

package_data = \
{'': ['*']}

install_requires = \
['detecta>=0.0.5,<0.0.6', 'numpy>=1.24.2,<2.0.0', 'pandas>=1.5.3,<2.0.0']

setup_kwargs = {
    'name': 'unfqt',
    'version': '0.2.0',
    'description': '',
    'long_description': 'None',
    'author': 'Joe Schr',
    'author_email': 'joe.vandelay@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# pragma pylint: disable=missing-docstring, invalid-name, pointless-string-statement
# isort: skip_file
# --- Do not remove these libs ---
from freqtrade.persistence import Trade
from freqtrade.strategy import stoploss_from_open
from freqtrade.enums import RunMode
from datetime import datetime
from technical.indicators import bollinger_bands, VIDYA
import talib.abstract as ta
import logging
import numpy as np  # noqa
import pandas as pd  # noqa
import pandas_ta as pta  # used as dataframe.ta
from pandas import DataFrame
from typing import Dict, List, Optional, Tuple, Union

from unfqt.helper_mixin import HelperMixin

# --------------------------------
# Import Strategy from different file


class CustomTakeprofitMixin(HelperMixin):
    custom_stoploss_config = {}

    def __init__(self, config: dict) -> None:
        super().__init__(config)

    # needs to be overridden
    def custom_takeprofit(self, pair: str, trade: 'Trade', current_time: datetime, current_rate: float, current_profit: float, **kwargs) -> float:
        """
        needs to be overriden
        """
        pass

    def is_takeprofit_reached(self, current_profit_pct, takeprofit_price_abs, open_rate_abs, **kwargs) -> bool:
        takeprofit_pct = takeprofit_price_abs / open_rate_abs - 1
        if(current_profit_pct >= takeprofit_pct):
            return True
        return False

    def stoploss_from_open(self, open_relative_stop: float, current_profit: float) -> float:
        return stoploss_from_open(open_relative_stop, current_profit)

    def stoploss_from_absolute(self, current_rate, stoploss_price):
        if stoploss_price >= current_rate:
            return 1
        return (stoploss_price / current_rate) - 1


class WaitForTakeProfitStoploss(CustomTakeprofitMixin):
    custom_stoploss_config = {
        'risk_reward_ratio': np.NaN,
        'atr_ratio': np.NaN,
        'min_rollback': np.NaN,
        'set_to_break_even_at_profit': np.NaN
    }
    stoploss = -0.30
    use_custom_stoploss = True

    def custom_stoploss(self, pair: str, trade: 'Trade', current_time: datetime,
                        current_rate: float, current_profit: float, **kwargs) -> float:
        """
        custom_stoploss using a risk/reward ratio
        """
        result = DO_NOTHING
        super().custom_stoploss(pair, trade, current_time, current_rate, current_profit)
        # get initial stoploss at trade.open_date
        trade_candle = self.get_trade_candle(trade)
        if not trade_candle.empty:
            if self.is_takeprofit_reached(current_profit,
                                          trade_candle['takeprofit_rate'], trade.open_rate):
                result = self.custom_takeprofit(pair, trade, current_time,
                                                current_rate, current_profit, **kwargs)

        return result

    def __init__(self, config: dict) -> None:
        super().__init__(config)


# set as close as possible to current price to trigger emergency sell on exchange
FORCE_SELL = 0.0000001
DO_NOTHING = 1  # no new stoploss gets set


class LockSellUntilTakeprofitMixin(WaitForTakeProfitStoploss):
    use_sell_signal = True

    def custom_takeprofit(self, pair: str, trade: 'Trade', current_time: datetime, current_rate: float, current_profit: float, **kwargs) -> float:
        result = DO_NOTHING
        current_candle = self.find_candle_datetime(
            current_time, trade.pair, now=current_time)
        if not current_candle.empty:
            should_sell = current_candle['sell'] > 0 if 'sell' in current_candle else False
            if(should_sell):
                result = FORCE_SELL
        return result

    def __init__(self, config: dict) -> None:
        super().__init__(config)


class SetTrailingStoplossMixin(LockSellUntilTakeprofitMixin):
    use_sell_signal = True

    def custom_takeprofit(self, pair: str, trade: 'Trade', current_time: datetime, current_rate: float, current_profit: float, **kwargs) -> float:
        result = super().custom_takeprofit(pair, trade, current_time,
                                           current_rate, current_profit, **kwargs)

        # check if we already got a sell signal
        if(abs(result) != DO_NOTHING):
            return result

        current_candle = self.find_candle_datetime(
            current_time, trade.pair, now=current_time)
        if not current_candle.empty:
            # calculate current trailing stoploss
            if 'trailing_stoploss_rate' in current_candle:
                result = self.stoploss_from_absolute(
                    current_rate, current_candle['trailing_stoploss_rate'])

        return result

    def __init__(self, config: dict) -> None:
        super().__init__(config)


class SetBreakEvenWhenTrailingStoplossIsNotSet(SetTrailingStoplossMixin):
    use_sell_signal = True

    def custom_takeprofit(self, pair: str, trade: 'Trade', current_time: datetime, current_rate: float, current_profit: float, **kwargs) -> float:
        result = super().custom_takeprofit(pair, trade, current_time,
                                           current_rate, current_profit, **kwargs)

        break_even_pct = (trade.fee_open + trade.fee_close)
        stoploss_break_even = self.stoploss_from_open(
            break_even_pct, current_profit)
        # break_even maybe still to big if takeprofit is very tiny
        # at this point return 0 would not set a new stoploss
        if(stoploss_break_even == 0):
            return result
        # super custom_takeprofit either did nothing
        # or set stoploss below break_even_pct
        # so we replace it with break_even_pct
        if(abs(result) == DO_NOTHING or abs(result) > abs(stoploss_break_even)):
            result = stoploss_break_even
        return result

    def __init__(self, config: dict) -> None:
        super().__init__(config)


class TakeprofitTrailingByATRMixin(CustomTakeprofitMixin):
    custom_stoploss_config = {
        'takeprofit_atr_ratio': 3,
        'min_rollback': 3,
        'trailing_atr_ratio': 1,
    }

    def populate_indicators(self, dataframe: DataFrame, metadata: dict) -> DataFrame:
        dataframe = super().populate_indicators(dataframe, metadata)
        dataframe['min'] = dataframe['low'].rolling(
            self.custom_stoploss_config['min_rollback']).min()
        dataframe['atr'] = dataframe.ta.atr()
        dataframe['takeprofit_rate'] = (
            dataframe['atr'] * self.custom_stoploss_config['takeprofit_atr_ratio']) + dataframe['low']
        dataframe['trailing_stoploss_rate'] = dataframe['min'] - \
            (dataframe['atr']*self.custom_stoploss_config['trailing_atr_ratio'])
        return dataframe

    def __init__(self, config: dict) -> None:
        super().__init__(config)


class TrailingByInversedPSARMixin(CustomTakeprofitMixin):
    custom_stoploss_config = {
        'takeprofit_atr_ratio': 3,
        'min_rollback': 3,
        'trailing_atr_ratio': 1,
        'risk_reward_ratio': 2
    }

    def populate_indicators(self, dataframe: DataFrame, metadata: dict) -> DataFrame:
        dataframe = super().populate_indicators(dataframe, metadata)
        if 'psar' not in dataframe.columns:
            dataframe['psar'] = ta.SAR(dataframe)

        dataframe['psar_inverse'] = dataframe.psar.where(~(dataframe['high'] < dataframe['psar']) & (
            dataframe.psar.shift() < dataframe.low.shift()), dataframe.psar.shift()-(dataframe.psar-dataframe.psar.shift()))
        dataframe['min'] = dataframe['low'].rolling(
            self.custom_stoploss_config['min_rollback']).min()
        dataframe['atr'] = dataframe.ta.atr()
        atr_trail = dataframe['min'] - \
            (dataframe['atr']*self.custom_stoploss_config['trailing_atr_ratio'])
        dataframe.psar_inverse.where(dataframe.psar_inverse <
                                     dataframe.low, atr_trail, inplace=True)
        dataframe.psar_inverse.ffill(inplace=True)
        dataframe['trailing_stoploss_rate'] = dataframe['psar_inverse']
        dataframe['stoploss_rate'] = dataframe['min'] - \
            (dataframe['psar_inverse'])
        dataframe['takeprofit_rate'] = ((dataframe['close']-dataframe['stoploss_rate'])
                                        * self.custom_stoploss_config['risk_reward_ratio'])+dataframe['close']
        return dataframe

    def __init__(self, config: dict) -> None:
        super().__init__(config)


DECAY_STOPLOSS = 0.30
